/*
 * modbus-utils - modbus.c
 *	version 0.1.0
 *
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <compiler.h>
#include <rejd_byteorder.h>
#include <string.h>

#define MODBUS_IOCTL_SET_BUS		_IO('g', 0x0)
#define MODBUS_IOCTL_SET_DEVICE		_IO('g', 0x1)
#define MODBUS_IOCTL_SET_RETRY_COUNT	_IO('g', 0x2)
#define MODBUS_IOCTL_SET_RAW		_IO('g', 0x3)
#define MODBUS_IOCTL_DETACH		_IO('g', 0x4)

#include <modbus.h>

#if CONFIG_DEBUG
#define debug(fmt...) fprintf(stderr, fmt)
#else
#define debug(fmt...) do { } while (0)
#endif

/**
 * modbus_init - initialize Modbus context
 * @ctx: context to initialize
 */
int modbus_init(struct modbus_context *ctx)
{
	ctx->fd = -1;
	ctx->emu = 0;
	INIT_LIST_HEAD(&ctx->reqs);
	return 0;
}

/**
 * modbus_set_emu - set modbus packet mode emulation
 * @ctx: context
 * @emu: use packet mode emulation
 */
int modbus_set_emu(struct modbus_context *ctx, int emu)
{
	ctx->emu = 1;
	return 0;
}

/**
 * modbus_open - open Modbus device that will be used for context
 * @ctx: context
 * @dev: path to Modbus device
 *
 * This function returns zero on success; negative error code otherwise.
 * In case of error previously opened device is not closed.
 */
int modbus_open(struct modbus_context *ctx, const char *dev)
{
	int fd;

	fd = open(dev, O_RDWR);
	if (fd == -1)
		return -errno;

	if (ctx->fd != -1)
		close(ctx->fd);
	ctx->fd = fd;
	return 0;
}

/**
 * modbus_set_packet - enable packet mode on specified context
 * @ctx: Modbus context
 */
int modbus_set_packet(struct modbus_context *ctx)
{
	int ret;

	if (ctx->emu)
		return 0;

	ret = ioctl(ctx->fd, MODBUS_IOCTL_DETACH, 0);
	if (ret)
		fprintf(stderr, "The kernel does not support "
				"MODBUS_IOCTL_DETACH.\n");
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_RAW, 1);
	if (ret)
		return -errno;
	return 0;
}

/**
 * modbus_request_init - initialize modbus requst
 * @r: request to initialize
 */
int modbus_request_init(struct modbus_request *r)
{
	r->c = NULL;
	r->c_len = 0;
	r->r = 0;
	r->r_len = 0;
	return 0;
}

/**
 * modbus_request_fill_id - generate unique request identifier
 * @r: request
 */
int modbus_request_fill_id(struct modbus_request *r)
{
	static int seq_id = 20;
	uint32_t *id;

	if (r->c_len < 4)
		return -EINVAL;

	id = (uint32_t *) r->c;
	r->id = seq_id++;
	*id = r->id;
	return 0;
}

/**
 * modbus_request_destroy - destroy Modbus request
 * @r: request to destroy
 */
int modbus_request_destroy(struct modbus_request *r)
{
	if (r->c != NULL)
		free(r->c);
	if (r->r != NULL)
		free(r->r);
	return 0;
}

/**
 * modbus_send - send request to specified context
 * @ctx: Modbus context
 * @r: request to send
 */
int modbus_send(struct modbus_context *ctx, struct modbus_request *r)
{
	int ret;

	r->ctx = ctx;
	if (!ctx->emu) {
		ret = write(ctx->fd, r->c, r->c_len);
		if (ret < 0)
			return ret;

		if (ret != (int) r->c_len)
			return -EAGAIN;
	}

	list_add(&r->list, &ctx->reqs);
	return 0;
}
	
/**
 * modbus_find_request - find request with specified ID
 * @ctx: Modbus context
 * @id: request id
 */
struct modbus_request *modbus_find_request(struct modbus_context *ctx,
		uint32_t id)
{
	struct modbus_request *r;

	list_for_each_entry(r, &ctx->reqs, list) {
		if (r->id == id)
			return r;
	}
	return NULL;
}

static int modbus_generate_error(struct modbus_request *r, unsigned int err)
{
	uint8_t *cmd = r->c;
	struct mbap_header *hdr;
	uint8_t *msg;

	hdr = xmalloc(sizeof(*hdr) + 2);
	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(1 + 2);

	msg = (void *)(hdr + 1);
	msg[0] = cmd[4] | 0x80;	/* error */
	msg[1] = err;
	r->r = hdr;
	r->r_len = sizeof(*hdr) + 2;
	return 0;
}
		
static int modbus_cmd_read(struct modbus_request *r)
{
	uint8_t *cmd = r->c;
	struct mbap_header *hdr;
	uint8_t *msg;
	uint16_t addr;
	uint16_t count;
	int ret;

	if (r->c_len < 4 + 5)
		return modbus_generate_error(r, MBEX_SLAVE_FAIL);

	addr = be16_to_cpu(get_unaligned((uint16_t *) (cmd + 5)));
	count = be16_to_cpu(get_unaligned((uint16_t *) (cmd + 7)));

	if (!count || (count > 0x7d))
		return modbus_generate_error(r, MBEX_INVAL);
	
	hdr = xmalloc(sizeof(*hdr) + 2 + 2 * count);
	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(1 + 2 + 2 * count);

	ret = pread(r->ctx->fd, (char *) (hdr + 1) + 2, 2 * count, 2 * addr);
	if (ret != 2 * count) {
		fprintf(stderr, "pread: readed %d, wanted %d\n", ret,
				2 * count);
		free(hdr);
		return modbus_generate_error(r, MBEX_FAULT);
	}
	msg = (void *)(hdr + 1);
	msg[0] = cmd[4];
	msg[1] = 2 * count;

	r->r = hdr;
	r->r_len = sizeof(*hdr) + 2 + 2 * count;
	return 0;
}

static int modbus_cmd_write_reg(struct modbus_request *r)
{
	uint8_t *cmd = r->c;
	struct mbap_header *hdr;
	uint8_t *msg;
	uint16_t addr;
	int ret;

	if (r->c_len < 4 + 5)
		return modbus_generate_error(r, MBEX_SLAVE_FAIL);

	addr = be16_to_cpu(get_unaligned((uint16_t *) (cmd + 5)));

	hdr = xmalloc(sizeof(*hdr) + 5);
	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(1 + 5);

	ret = pwrite(r->ctx->fd, (char *) (cmd + 7), 2, 2 * addr);
	if (ret != 2) {
		fprintf(stderr, "pwrite: written %d, wanted %d\n", ret, 2);
		free(hdr);
		return modbus_generate_error(r, MBEX_FAULT);
	}
	msg = (void *)(hdr + 1);
	memcpy(msg, cmd + 4, 5);

	r->r = hdr;
	r->r_len = sizeof(*hdr) + 5;
	return 0;
}

static int modbus_cmd_write(struct modbus_request *r)
{
	uint8_t *cmd = r->c;
	struct mbap_header *hdr;
	uint8_t *msg;
	uint16_t addr;
	uint16_t count;
	int ret;

	if (r->c_len < 4 + 6)
		return modbus_generate_error(r, MBEX_SLAVE_FAIL);

	addr = be16_to_cpu(get_unaligned((uint16_t *) (cmd + 5)));
	count = be16_to_cpu(get_unaligned((uint16_t *) (cmd + 7)));

	if (!count || (count > 0x7b))
		return modbus_generate_error(r, MBEX_INVAL);

	if ((int) r->c_len < 4 + 6 + cmd[9])
		return modbus_generate_error(r, MBEX_SLAVE_FAIL);

	if (cmd[9] < 2 * count)
		return modbus_generate_error(r, MBEX_SLAVE_FAIL);


	hdr = xmalloc(sizeof(*hdr) + 5);
	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(1 + 5);

	ret = pwrite(r->ctx->fd, (char *) (cmd + 10), 2 * count, 2 * addr);
	if (ret != 2 * count) {
		fprintf(stderr, "pwrite: written %d, wanted %d\n", ret,
				2 * count);
		free(hdr);
		return modbus_generate_error(r, MBEX_FAULT);
	}
	msg = (void *)(hdr + 1);
	memcpy(msg, cmd + 4, 5);

	r->r = hdr;
	r->r_len = sizeof(*hdr) + 5;
	return 0;
}

static int modbus_process_emu(struct modbus_context *ctx,
		struct modbus_request **re)
{
	struct modbus_request *r;
	uint8_t *cmd;

	if (list_empty(&ctx->reqs))
		return -EAGAIN;

	r = list_entry(ctx->reqs.next, struct modbus_request, list);
	list_del(&r->list);

	if (r->c_len < 5)
		modbus_generate_error(r, MBEX_SLAVE_FAIL);

	cmd = r->c;
	switch (cmd[4]) {
	case 3:
	case 4:
		modbus_cmd_read(r);
		break;
	case 6:
		modbus_cmd_write_reg(r);
		break;
	case 0x10:
		modbus_cmd_write(r);
		break;
	default:
		modbus_generate_error(r, MBEX_NOSYS);
	}

	*re = r;
	return 0;
}

int modbus_process(struct modbus_context *ctx, struct modbus_request **re,
		int events)
{
	struct modbus_request *r;
	struct mbap_header *hdr;
	uint32_t id;
	void *d;
	int ret;

	if (ctx->emu)
		return modbus_process_emu(ctx, re);

	d = malloc(MBAP_MTU);
	if (d == NULL)
		return -ENOMEM;

	ret = read(ctx->fd, (char *) d + sizeof(*hdr) - 4, MBAP_MTU);
	if (ret < 0) {
		perror("read");
		goto err_free;
	}

	if (ret < 4) {
		ret = -EPROTO;
		goto err_free;
	}
	id = get_unaligned((uint32_t *) ((char *) d + sizeof(*hdr) - 4));

	r = modbus_find_request(ctx, id);
	if (r == NULL) {
		debug("modbus: unknown id: %d", id);
		goto err_free;
	}
	r->r_len = ret + sizeof(*hdr) - 4;
	r->r = d;
	hdr = d;

	hdr->proto = cpu_to_be16(0);
	hdr->len = cpu_to_be16(r->r_len - 6);

	list_del(&r->list);

	if (re != NULL)
		*re = r;
	return 0;

err_free:
	free(d);
	return ret;
}

int modbus_poll(struct modbus_context *ctx, struct modbus_request **r,
		int timeout)
{
	return -ENOSYS;
}

int modbus_set_bus(struct modbus_context *ctx, unsigned int bus)
{
	int ret;

	if (ctx->emu)
		return 0;

	if (ctx->fd == -1)
		return -EBADF;
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_BUS, bus);
	if (ret)
		return -errno;
	return 0;
}

int modbus_set_device(struct modbus_context *ctx, unsigned int device)
{
	int ret;

	if (ctx->emu)
		return 0;

	if (ctx->fd == -1)
		return -EBADF;
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_DEVICE, device);
	if (ret)
		return -errno;
	return 0;
}

int modbus_set_addr(struct modbus_context *ctx, const struct modbus_addr *addr)
{
	int ret;

	if (ctx->fd == -1)
		return -EBADF;
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_BUS, addr->bus);
	if (ret)
		return -errno;
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_DEVICE, addr->addr);
	if (ret)
		return -errno;
	return 0;
}

int modbus_set_retry(struct modbus_context *ctx, unsigned int retry)
{
	int ret;

	if (ctx->emu)
		return 0;

	if (ctx->fd == -1)
		return -EBADF;
	ret = ioctl(ctx->fd, MODBUS_IOCTL_SET_RETRY_COUNT, retry);
	if (ret)
		return -errno;
	return 0;
}

int modbus_read(struct modbus_context *ctx, struct modbus_addr *addr,
		void *buf, unsigned int len, unsigned int offset)
{
	int ret;

	if (ctx->fd == -1)
		return -EBADF;

	ret = modbus_set_addr(ctx, addr);
	if (ret)
		return ret;
	ret = pread(ctx->fd, buf, len, offset);
	if (ret == -1)
		return -errno;
	return ret;
}

int modbus_close(struct modbus_context *ctx)
{
	if (ctx->fd != -1)
		close(ctx->fd);
	return 0;
}

int modbus_destroy(struct modbus_context *ctx)
{
	modbus_close(ctx);
	return 0;
}
